/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil -*- */
/*
 * Copyright (c) 2013-2014 The University of Tennessee and The University
 *                         of Tennessee Research Foundation.  All rights
 *                         reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <math.h>
#include <mpi.h>

/* Performs a comm_split, returns uniformely MPI_ERR_PROC_FAILED or
 * MPI_SUCCESS */
int MPIX_Comm_split_safe(MPI_Comm comm, int color, int key, MPI_Comm *newcomm) {
    int rc;
    int flag;

    rc = MPI_Comm_split(comm, color, key, newcomm);
    flag = (MPI_SUCCESS==rc);	
    OMPI_Comm_agree( comm, &flag);
    if( !flag ) {
        if( rc == MPI_SUCCESS ) {
            MPI_Comm_free( newcomm );
            rc = MPI_ERR_PROC_FAILED;
        }
    }
    return rc;
}

void print_timings( MPI_Comm scomm, double tff, double twf );
int rank, verbose=0; /* makes this global (for printfs) */
#define COUNT 1024

int main( int argc, char* argv[] ) { 
    MPI_Comm fcomm, ncomm; /* a comm to work, and a comm to create */
    int np;
    int rc; /* error code from MPI functions */
    char estr[MPI_MAX_ERROR_STRING]=""; int strl; /* error messages */
    double start, tff, twf; /* timings */
    double array[COUNT];
    
    MPI_Init( &argc, &argv );
    MPI_Comm_size( MPI_COMM_WORLD, &np );
    MPI_Comm_rank( MPI_COMM_WORLD, &rank );
    if( !strcmp( argv[argc-1], "-v" ) ) verbose=1;

    /* Let's work on a copy of MPI_COMM_WORLD, good practice anyway. */
    MPI_Comm_dup( MPI_COMM_WORLD, &fcomm );
    /* We set an errhandler on fcomm, so that a failure is not fatal anymore. */
    MPI_Comm_set_errhandler( fcomm, MPI_ERRORS_RETURN );

    /* Do a first non FT split, and time it */
    start=MPI_Wtime();
    rc = MPI_Comm_split( fcomm, 0, 0, &ncomm );
    tff=MPI_Wtime()-start;
    if(verbose) { 
        MPI_Error_string( rc, estr, &strl );
        printf( "Rank %04d: Split (unsafe) completed (rc=%s) duration %g (s)\n", rank, estr, tff );
    }
    MPI_Comm_free( &ncomm );

    /* Time a consistently creating split */
    start=MPI_Wtime();
    rc = MPIX_Comm_split_safe( fcomm, 0, 0, &ncomm );
    twf=MPI_Wtime()-start;
    if(verbose) { 
        MPI_Error_string( rc, estr, &strl );
        printf( "Rank %04d: Split_safe completed (rc=%s) duration %g (s)\n", rank, estr, twf );
    }
    MPI_Comm_free( &ncomm );

    print_timings( fcomm, tff, twf );

    MPI_Comm_free( &fcomm );

    MPI_Finalize();
    return EXIT_SUCCESS;
}

void print_timings( MPI_Comm scomm, 
                    double tff, 
                    double twf ) {
    /* Storage for min and max times */
    double mtff, Mtff, mtwf, Mtwf;

    MPI_Reduce( &tff, &mtff, 1, MPI_DOUBLE, MPI_MIN, 0, scomm );
    MPI_Reduce( &tff, &Mtff, 1, MPI_DOUBLE, MPI_MAX, 0, scomm );
    MPI_Reduce( &twf, &mtwf, 1, MPI_DOUBLE, MPI_MIN, 0, scomm );
    MPI_Reduce( &twf, &Mtwf, 1, MPI_DOUBLE, MPI_MAX, 0, scomm );

    if( 0 == rank ) printf( 
        "## Timings ########### Min         ### Max         ##\n"
        "Split   (unsafe)    # %13.5e # %13.5e\n"
        "Split   (safe)      # %13.5e # %13.5e\n"
        , mtff, Mtff, mtwf, Mtwf );   
}

